mod cli;
mod csv;
mod packet;
#[cfg(feature = "tdl")]
mod tdl;
mod utils;

use std::{
    fs::File,
    io::{self, Read},
    time::{Duration, Instant},
};

use clap::Parser;
use log::{debug, error, info, warn};
use miette::{Context, IntoDiagnostic, Result};
use serialport::SerialPort;
use skyward_mavlink::{
    lyra::MavMessage,
    mavlink::{write_v1_msg, MavHeader},
};

#[cfg(feature = "tdl")]
use crate::cli::ModeFile;
use crate::{
    cli::Cli,
    csv::read_packets_from_csv,
    packet::PacketSequence,
    utils::{get_first_stm32_serial_port, print_stats},
};

const ACK: u8 = 0x06;

fn main() -> Result<()> {
    let args: Cli = Cli::parse();

    let lvl = if args.verbose {
        log::LevelFilter::Debug
    } else {
        log::LevelFilter::Info
    };

    // initialize logger
    simple_logger::SimpleLogger::new()
        .with_level(lvl)
        .with_local_timestamps()
        .env()
        .init()
        .into_diagnostic()?;

    if !args.no_sync {
        info!("SYNC mode ENABLED");
    } else {
        info!("SYNC mode DISABLED");
    }

    let packets = get_packets(&args)?;
    debug!("collected {} packets", packets.len());

    // convert to packet sequence data structure
    let mut packets = PacketSequence::from(packets);

    // Open the first serialport available.
    let port_name = if let Some(name) = args.port {
        name
    } else if let Some(name) = get_first_stm32_serial_port()? {
        debug!("found STM32 device on {}", name);
        name
    } else {
        error!("FATAL: no STM32 serial port found, your device may not be connected properly");
        std::process::exit(1);
    };
    debug!("connecting to serial port {}...", port_name);
    let mut port = serialport::new(port_name, args.baud_rate)
        .open()
        .into_diagnostic()
        .wrap_err("Failed to open serial port")?;
    let mut write_port = port.try_clone().into_diagnostic()?;

    // read from buffer and print to stdout
    let mut buffer = [0u8; 1000];
    let mut start: Option<Instant> = None;
    let mut times = Vec::new();

    // send packet if there is one
    while let Some(packet) = packets.wait_next() {
        if let Some(t) = start {
            times.push(t.elapsed().as_millis());
        }
        start = Some(Instant::now());
        write_port.write_packet(packet.to_owned())?;
        debug!(
            "sent packet ({}/{})",
            packets.sent_packets(),
            packets.total_packets()
        );

        // if sync mode is enabled then wait for ACK
        if !args.no_sync {
            wait_for_ack(&mut port, &mut buffer);
        }
    }

    info!("sent {} packets", packets.sent_packets());
    print_stats(times);

    Ok(())
}

#[cfg(feature = "tdl")]
fn get_packets(args: &Cli) -> Result<Vec<MavMessage>> {
    let packets: Vec<MavMessage> = match &args.mode {
        ModeFile::Csv { csv_file } => {
            info!("reading from csv file {}", csv_file.display());
            read_packets_from_csv(File::open(csv_file).into_diagnostic()?).into_diagnostic()?
        }
        ModeFile::Tdl {
            tdl_file,
            lat,
            lon,
            alt,
        } => {
            info!("reading from tdl file {}", tdl_file.display());
            tdl::generate_packets_from_tdl(tdl_file, *lat, *lon, *alt)?
        }
    };
    debug!("collected {} packets", packets.len());
    Ok(packets)
}

#[cfg(not(feature = "tdl"))]
fn get_packets(args: &Cli) -> Result<Vec<MavMessage>> {
    info!("reading from csv file {}", args.csv_file.display());
    let packets =
        read_packets_from_csv(File::open(&args.csv_file).into_diagnostic()?).into_diagnostic()?;
    debug!("collected {} packets", packets.len());
    Ok(packets)
}

/// Wait for ACK signal from the device
fn wait_for_ack(port: &mut Box<dyn SerialPort>, buffer: &mut [u8]) {
    // wait ACK
    loop {
        match port.read(buffer) {
            Ok(t) => {
                // eprint!("{}", String::from_utf8_lossy(&buffer[..t]));
                // if didn't get the signal then print the buffer and exit with error
                if buffer[..t].contains(&ACK) {
                    break;
                } else {
                    warn!("didn't get the signal");
                    std::thread::sleep(Duration::from_millis(10));
                }
            }
            Err(ref e) if e.kind() == io::ErrorKind::TimedOut => (),
            Err(e) => {
                error!("FATAL: {:?}", e);
                std::process::exit(1);
            }
        }
    }
    // eprintln!();
}

trait MavLinkPort {
    fn write_packet(&mut self, packet: MavMessage) -> Result<()>;
}

impl<T: AsRef<dyn SerialPort> + std::io::Write> MavLinkPort for T {
    fn write_packet(&mut self, packet: MavMessage) -> Result<()> {
        let header = MavHeader::default();
        write_v1_msg(self, header, &packet)
            .into_diagnostic()
            .wrap_err("Failed to write mavlink message to serial port")?;
        Ok(())
    }
}
