use std::{path::Path, time::Instant};

use miette::{miette, Context, IntoDiagnostic, Result};
use skyward_mavlink::lyra::{MavMessage, ROCKET_FLIGHT_TM_DATA};
use tdl_core::{
    semantics::{Executor, Point, Trajectory},
    syntax::CachedLexer,
};

pub fn generate_packets_from_tdl(
    tdl_file: impl AsRef<Path>,
    gps_lat: f64,
    gps_lon: f64,
    gps_alt: f64,
) -> Result<Vec<MavMessage>> {
    let tdl_content = std::fs::read_to_string(tdl_file.as_ref())
        .into_diagnostic()
        .wrap_err("Failed to read TDL file")?;
    let program = CachedLexer::new(&tdl_content)
        .parse_program()
        .wrap_err("Failed to parse TDL program")?;
    let traj = Executor::new(program)
        .run()
        .or(Err(miette!("Failed to execute TDL program")))?;
    let generator = RocketFlightGenerator::new(gps_lat, gps_lon, gps_alt);
    generator.generate_packets(traj)
}

trait MessageGenerator {
    fn generate_packets(self, traj: Trajectory) -> Result<Vec<MavMessage>>;
}

struct RocketFlightGenerator {
    normalization: Option<u128>,
    max_altitude: f64,
    initial_gps: (f64, f64, f64),
    last_time: Option<f64>,
    last_pos: Option<(f64, f64, f64)>,
    last_vel: Option<(f64, f64, f64)>,
    last_acc: Option<(f64, f64, f64)>,
    ada_flip: bool,
}

impl RocketFlightGenerator {
    fn new(lat: f64, lon: f64, alt: f64) -> Self {
        Self {
            initial_gps: (lat, lon, alt),
            normalization: None,
            max_altitude: 0.0,
            last_time: None,
            last_pos: None,
            last_vel: None,
            last_acc: None,
            ada_flip: false,
        }
    }

    fn compute_velocity(&self, point: Point) -> Option<(f64, f64, f64)> {
        if let Some((x, y, z)) = self.last_pos {
            let dt = point.time - self.last_time?;
            let dx = point.x - x;
            let dy = point.y - y;
            let dz = point.z - z;
            Some((dx / dt, dy / dt, dz / dt))
        } else {
            None
        }
    }

    fn compute_acceleration(
        &self,
        vel: Option<(f64, f64, f64)>,
        time: f64,
    ) -> Option<(f64, f64, f64)> {
        let (vx, vy, vz) = vel?;
        if let Some((lvx, lvy, lvz)) = self.last_vel {
            let dt = time - self.last_time?;
            let dvx = vx - lvx;
            let dvy = vy - lvy;
            let dvz = vz - lvz;
            Some((dvx / dt, dvy / dt, dvz / dt))
        } else {
            None
        }
    }

    fn consume_point(&mut self, point: Point) -> ROCKET_FLIGHT_TM_DATA {
        let Point { x, y, z, time: t } = point;

        // normalize time relative to now
        let time = if let Some(offset) = self.normalization {
            (t * 1e6).floor().min(u128::MAX as f64) as u128 + offset
        } else {
            let time_us = (t * 1e6).floor().min(u128::MAX as f64) as u128;
            self.normalization = Some(Instant::now().elapsed().as_micros() - time_us);
            0
        };

        // check for overflow
        if time > u64::MAX as u128 {
            panic!("Time overflow");
        }

        // calculate velocity and acceleration
        let vel = self.compute_velocity(point);
        let acc = self.compute_acceleration(vel, t).unwrap_or_default();
        let vel = vel.unwrap_or_default();
        // and save them
        self.last_time = Some(t);
        self.last_pos = Some((x, y, z));
        self.last_vel = Some(vel);
        self.last_acc = Some(acc);

        // check for max altitude reached
        if z >= self.max_altitude && !self.ada_flip {
            self.max_altitude = z;
        } else {
            self.ada_flip = true;
        }

        // deconstruct the tuple
        let (x, y, z) = (x as f32, y as f32, z as f32);
        let (vx, vy, vz) = (vel.0 as f32, vel.1 as f32, vel.2 as f32);
        let (ax, ay, az) = (acc.0 as f32, acc.1 as f32, acc.2 as f32);

        let pin = if self.ada_flip { 1 } else { 0 };

        let (lat, lon, alt) = self.initial_gps;
        let (lat, lon, alt) = (lat as f32, lon as f32, alt as f32);

        ROCKET_FLIGHT_TM_DATA {
            timestamp: time as u64,
            pressure_digi: 88400.0,
            pressure_static: 88400.0,
            airspeed_pitot: 10.0,
            altitude_agl: z,
            acc_x: ax,
            acc_y: ay,
            acc_z: az,
            gyro_x: 0.0,
            gyro_y: 0.0,
            gyro_z: 0.0,
            mag_x: 0.0,
            mag_y: 0.0,
            mag_z: 0.0,
            gps_lat: lat,
            gps_lon: lon,
            gps_alt: alt,
            nas_n: y,
            nas_e: x,
            nas_d: -z,
            nas_vn: vy,
            nas_ve: vx,
            nas_vd: -vz,
            nas_qx: 0.0,
            nas_qy: 0.0,
            nas_qz: 0.0,
            nas_qw: 1.0,
            nas_bias_x: 0.0,
            nas_bias_y: 0.0,
            nas_bias_z: 0.0,
            battery_voltage: 8.0,
            cam_battery_voltage: 8.0,
            current_consumption: 0.0,
            temperature: 40.0,
            fmm_state: 0,
            nas_state: 0,
            gps_fix: 3,
            pin_nosecone: pin,
            cutter_presence: 0,
            logger_error: 0,
            pressure_ada: 0.0,
            pressure_dpl: 0.0,
            ada_vert_speed: 0.0,
            mea_mass: 0.0,
            abk_angle: 0.0,
            pin_quick_connector: 0.0,
            ada_state: 0,
            dpl_state: 0,
            abk_state: 0,
            mea_state: 0,
            pin_launch: 0,
            pin_expulsion: 0,
        }
    }
}

impl MessageGenerator for RocketFlightGenerator {
    fn generate_packets(mut self, traj: Trajectory) -> Result<Vec<MavMessage>> {
        let Trajectory { points } = traj;
        let mut packets = Vec::new();
        for point in points {
            let packet = self.consume_point(point);
            packets.push(MavMessage::ROCKET_FLIGHT_TM(packet));
        }
        Ok(packets)
    }
}
